// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	EnergyCorrector --- A class EnergyCorrector is defined
//
//	DESCRIPTION:
//		Instances of class EnergyCorrector can
//			a) offset correct PHA channels
//			b) gain correct offset-corrected PHA channels (PHA->PI conversion)
//
//	CHANGE HISTORY:
//		UL, 1999-05-25: 1.0.0: - first version from CalAtomsCommon
//		UL, 1999-07-09: 1.1.0: - added RgsEnergyCorrector which implements
//								 RGS-specific offset correction
//      UL, 1999-08-13: 1.2.0: - added piTokeV()
//		UL, 1999-09-13: 1.3.0: - changed piTokeV() -> piToeV
//		UL, 1999-12-16: 1.4.0: - added special EmosEnergyCorrector
//		UL, 2000-02-02: 1.5.0: - added qSplitRatio() to retrieve split ratios
//								 for RGS
//		UL, 2000-02-29: 1.6.0: - brought RGS offset I/F in line with [2]
//		UL, 2000-05-09: 1.7.0: - new EMOS gain correction (depends on pattern)
//		UL, 2001-04-03: 1.7.1: - couple of transparent changes to implement
//								 ALGOID-1 RGS offset correction
//		UL, 2001-10-30: 1.8.0: - RGS HTR mode offset coorection method added
//		UL, 2001-12-04: 1.9.0: - added EnergyCorrector::badPixelTableCode()
//		UL, 2002-07-17: 2.0.0: - added three pn-specific methods:
//									EnergyCorrector::additionalGainCtiCorrect()
//									EnergyCorrector::gainTimingCorrect()
//									EnergyCorrector::gainBurstCorrect()
//		UL, 2002-11-18: 2.1.0: - EnergyCorrector::gainTiming/BurstCorrect()
//								 were not declared virtual
//
//	AUTHOR:
//		U. Lammers, Mon Oct  6 15:32:57 WET DST 1997
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Calibration Access and Data Handbook, C. Erd et al.
//
//	$Header: /sasbuild/sas_git_repo/cvs_copy/sasdev/sas/packages/cal/src/EnergyCorrector.h,v 1.38 2014/03/19 15:50:24 rsaxton Exp $
//

#ifndef ENERGY_CORRECTOR_H
#define ENERGY_CORRECTOR_H

#include "Ccf.h"
#include "CalChild.h"
#include "CalAtomsBase.h"
#include "CalDataVector.h"
#include "slatec++.h"
#include "STime.h"
#include "Image.h"

#include <vector>

//
//	---( EnergyCorrector )--------------------------------------------------
//
class EnergyCorrector : public CalAtom,
			  public virtual CalChild<CalAtomCollection<EnergyCorrector> > {
public:
	EnergyCorrector(Ccd ccd,
					const CcfConstituentTable &aduConv,
					CalAtomCollection<EnergyCorrector> *parent = 0);
		//
		//	PURPOSE:
		//		Construct EnergyCorrector object
		//
		//	ARGUMENTS:
		//		aduConv	: CCF table containing data for generic offset/gain
		//				  correction
		//		ccd		: CCD identifier
		//


	static unsigned id() { return Ccf::ADUConv; }
		//
		//	PURPOSE:
		//		Return a unique ID which for mapping Cal atoms onto
		//		CCF constituents
		//


	static const string INVALID_BAD_PIXEL_TABLE_CODE;
	const string &badPixelTableCode() const { return _bptCode; }
		//
		//	PURPOSE:
		//		Return Bad-Pixel-Table code read from attribute
		//		*ADUCONV*:OFFSET_GAIN%BPT_CODE; returns
		//		EnergyCorrector::INVALID_BPT_CODE if not present
		//		

	virtual void dump(std::ostream &os);
		//
		//	PURPOSE:
		//		Dump data members to an ostream
		//
		//	ARGUMENTS:
		//		In:
		//			os:	stream to print on
		//


	float offsetValue(Node node) const;
		//
		//	RETURN:
		//		Fixed offset value for node <node>
		//


	virtual CalReal32Vector &offsetCorrect(const CalReal64Vector &time,
										   const CalReal32Vector &eIn,
										   CalReal32Vector &eOut) const;
	virtual CalReal32Vector &offsetCorrect(const CalReal64Vector &time,
										   const CalInt8Vector &nodes,
										   const CalReal32Vector &eIn,
										   CalReal32Vector &eOut) const;
	virtual CalReal32Vector &offsetCorrect(const CalReal64Vector &time,
										   const CalInt16Vector &rawX,
										   const CalInt8Vector &nodes,
										   const CalInt16Vector &eIn,
										   CalReal32Vector &eOut) const;
		//
		//	PURPOSE:
		//		Correct a vector of PHA channels for offset (time dependency
		//		currently ignored):
		//			eOut = eIn - offset
		//		Derived classes implement concrete versions
		//
		//	ARGUMENTS:
		//		In:
		//			time	: vector with time tags in TIMCOORD frame
		//					  (currently ignored)
		//			eIn		: vector with PHA channel value to correct
		//			nodes	: node vector with nodes identifiers
		//					  in the range [0, MAX_NODES-1] signifying the
		//					  readout node which gave the pha channel
		//					  value passed via `eIn'; if omitted by using
		//					  the first form, the node is assumed to be
		//					  fixed and taken from the corresponding CAL
		//					  state variable
		//		Out:
		//			eOut	: offset corrected PHA channel vector
		//


	virtual CalReal32Vector &gainCorrect(const CalReal64Vector &time,
										 const CalReal32Vector &eIn,
										 CalReal32Vector &eOut) const;
	virtual CalReal32Vector &gainCorrect(const CalReal64Vector &time,
										 const CalInt8Vector &nodes,
										 const CalReal32Vector &eIn,
										 CalReal32Vector &eOut) const;
		//
		//	PURPOSE:
		//		Convert offset corrected PHA channels to PI, i.e., apply
		//		(time-dependent) gain correction
		//
		//	ARGUMENTS:
		//		In:
		//			time	: vector with time tags in TIMCOORD frame
		//					  (currently ignored)
		//			nodes	: see offsetCorrect() above
		//			eIn		: vector with pha channels to correct
		//		Out:
		//			eOut	: vector with gain corrected PI channels
		//
		//	RETURNS:
		//		vector with gain corrected PI channels
		//


	virtual CalReal32Vector &piToeV(const CalReal32Vector &eIn,
									CalReal32Vector &eOut) const;
		//
		//	PURPOSE:
        //      Convert a vector of PI channels to eV:
		//			eOut = eIn * 1.0
		//
		//	ARGUMENTS:
		//		In:
		//			eIn		: vector with PI channel values to convert
		//		Out:
		//			eOut	: vector with eV values
		//

	virtual CalReal32Vector &eVToPi(const CalReal32Vector &eIn,
									CalReal32Vector &eOut) const;
		//
		//	PURPOSE:
        //      Convert a vector of eV energies to PI channels 
		//			eOut = eIn * 1.0
		//
		//	ARGUMENTS:
		//		In:
		//			eIn		: vector with PI channel values to convert
		//		Out:
		//			eOut	: vector with eV values
		//


	virtual ~EnergyCorrector() {}
		//
		//	PURPOSE:
		//		Destruct EnergyCorrector object
		//


	//
	//	--- I/F for specialized versions of above generic operations
	//

	//
	//	EMOS
	//
	virtual CalReal32Vector &gainCorrect(const CalReal64Vector &,
										 const CalReal32Vector &,
										 const CalInt8Vector &,
										 CalReal32Vector &eOut) const
		{
			errstr<<"Generic energy corrector cannot perform EMOS gain "
					"correction"
				  <<Sas::error("UnsuitableEnergyCorrector");
		  return eOut;
		}

	virtual std::vector<float> e1Coefficients() const
		{ return std::vector<float>(); }
	virtual std::vector<float> e2Coefficients() const
		{ return std::vector<float>(); }
	virtual std::vector<float> e3Coefficients() const
		{ return std::vector<float>(); }
	virtual std::vector<float> e4Coefficients() const
		{ return std::vector<float>(); }


	//
	//	EPN
	//
	virtual CalReal32Vector &gainCorrect(const CalReal64Vector &,
										 const CalInt16Vector &,
										 const CalInt16Vector &,
										 const CalReal32Vector &,
										 CalReal32Vector &eOut) const
		{
			errstr<<"Generic energy corrector does not implement EPN-specific "
					"gain correction"
				  <<Sas::error("UnsuitableEnergyCorrector");
			return eOut;
		}

	virtual std::vector<double> additionalGainCtiCorrect(
										CalReal32Vector &,
										const CalInt16Vector &,
										double,
										bool, bool, bool, bool, bool) const
		{
			errstr<<"Generic energy corrector does not implement additional "
					"pn gain/CTI corrections"
				  <<Sas::error("UnsuitableEnergyCorrector");
			return std::vector<double>();
		}

	virtual void spatialCtiCorrect(CalReal32Vector &,
                                       const CalInt16Vector &,
                                       const CalInt16Vector &) const
		{
			errstr<<"Generic energy corrector does not implement pn spatial "
					"CTI correction"
				  <<Sas::error("UnsuitableEnergyCorrector");
		}

	virtual void spatialCtiCorrectSD20(CalReal32Vector &,
                                       const CalInt16Vector &,
                                       const CalInt16Vector &) const
		{
			errstr<<"Generic energy corrector does not implement pn spatial "
					"CTI correction"
				  <<Sas::error("UnsuitableEnergyCorrector");
		}

	virtual void bkgDepGainCorrect(CalReal32Vector &,
                                       int32,
                                       int16,
				       int16) const
		{
			errstr<<"Generic energy corrector does not implement pn bckgnd "
					"dependent gain correction"
				  <<Sas::error("UnsuitableEnergyCorrector");
		}

	virtual void combinedEventCorrect(CalReal32Vector &,
                                       const CalInt16Vector &) const
		{
			errstr<<"Generic energy corrector does not implement pn combined event "
					"CTI correction"
				  <<Sas::error("UnsuitableEnergyCorrector");
		}

	virtual void gainTimingCorrect(CalReal32Vector &) const
		{
			errstr<<"Generic energy corrector does not implement pn gain "
					"timing correction"
				  <<Sas::error("UnsuitableEnergyCorrector");
		}

	virtual void gainBurstCorrect(CalReal32Vector &) const
		{
			errstr<<"Generic energy corrector does not implement pn gain "
					"burst correction"
				  <<Sas::error("UnsuitableEnergyCorrector");
		}

	virtual void gainEffCorrect(CalReal32Vector &) const
		{
			errstr<<"Generic energy corrector does not implement pn gain "
					"extended full frame correction"
				  <<Sas::error("UnsuitableEnergyCorrector");
		}

	virtual void gainFfCorrect(CalReal32Vector &) const
		{
			errstr<<"Generic energy corrector does not implement pn gain "
					"full frame correction"
				  <<Sas::error("UnsuitableEnergyCorrector");
		}

	virtual double reEmissionThreshold(double, int16) const
		{
			errstr<<"Generic energy corrector does not implement re-emission "
					"threshold algorithm"
				  <<Sas::error("UnsuitableEnergyCorrector");
			return 0.;
		}

	virtual std::vector<double> getRateDepPhaCoeffs(double, bool) const
                {
                        errstr<<"Generic energy corrector does not implement rate-Dependent PHA correction "
                                        "algorithm"
                                  <<Sas::error("UnsuitableEnergyCorrector");
			return std::vector<double>();
                }

	virtual std::vector<float> getRateDepGainCoeffs() const
                {
                        errstr<<"Generic energy corrector does not implement rate-Dependent CTI correction "
                                        "algorithm"
                                  <<Sas::error("UnsuitableEnergyCorrector");
			return std::vector<float>();
                }

	virtual void readRDPHAcoeffs(Table*) 
                {
                        errstr<<"Generic energy corrector does not implement rate-Dependent PHA correction "
                                        "algorithm"
                                  <<Sas::error("UnsuitableEnergyCorrector");
                }

	//
	//	RGS
	//
	virtual CalReal32Vector &offsetCorrect(const CalReal64Vector &, //time
										   const CalInt16Vector &,	//rawX
										   const CalInt16Vector &,	//rawY
										   const CalInt8Vector &,	//shape
										   const CalInt16Vector &,	//eIn
										   CalReal32Vector &eOut) const
		{
			errstr<<"Generic energy corrector cannot perform RGS offset "
					"correction"
				  <<Sas::error("UnsuitableEnergyCorrector");
		  return eOut;
		}

	virtual CalReal32Vector &offsetCorrect(const CalReal64Vector &,	//time
										   const CalInt16Vector &,	//rawX
										   const CalInt16Vector &,	//rawY
										   const CalInt8Vector &,	//shape
										   const CalInt8Vector &,	//nodes
										   const CalInt16Vector &,	//eIn
										   CalReal32Vector &eOut) const
		{
			errstr<<"Generic energy corrector cannot perform RGS offset "
					"correction"
				  <<Sas::error("UnsuitableEnergyCorrector");
		  return eOut;
		}


	virtual double rejectionThreshold(unsigned) const
		{
			errstr<<"Generic energy corrector has no knowledge about "
					"rejection/acceptance thresholds"
				  <<Sas::error("UnsuitableEnergyCorrector");
		  return 0.;
		}
	virtual double rejectionThreshold(unsigned, Node) const
		{ return rejectionThreshold(0); }

	virtual double acceptanceThreshold(unsigned) const
		{ return rejectionThreshold(0); }
	virtual double acceptanceThreshold(unsigned, Node) const
		{ return rejectionThreshold(0); }

	virtual unsigned nGrades() const { return 0; }
	virtual std::vector<float> qSplitRatios() const
		{ return std::vector<float>(); }

protected:
	//	data members
	std::vector<float>	_offsetConst, _offsetSlope, _gain;
    double				_eVperPi, _eV0; //  eV per PI channel value
	STime				_offsetRefTime;

	std::string			_bptCode;	//	bad-pixel-table code from CCF
};



//
//	a version of the EnergyCorrector specialized for EMOS
// 
class EmosEnergyCorrector : public virtual EnergyCorrector {
public:
	EmosEnergyCorrector(Ccd ccd,
						const CcfConstituentTable &aduConv,
						const Table *eCombine,
						CalAtomCollection<EnergyCorrector> *parent = 0);
		//
		//	PURPOSE:
		//		Construct EmosEnergyCorrector object
		//
		//	ARGUMENTS:
		//		Ccd			: CCD number this EnergyCorrector is applicable to
		//		aduConv		: CCF AduConv table from which to initialize object
		//		eCombine	: CCF table with energy combination coefficients
		//		parent		: parent of the object
		//


	CalReal32Vector &gainCorrect(const CalReal64Vector &time,
								 const CalReal32Vector &eIn,
								 const CalInt8Vector &pattern,
								 CalReal32Vector &eOut) const;
		//
		//	PURPOSE:
		//		Perform EMOS gain correction
		//
		//	ARGUMENTS:
		//		time	: vector with time tags
		//		pattern	: vector with pattern ID [0, 31]
		//		eIn		: vector with uncorrected raw ADUs
		//		eOut	: gain corrected ADUs
		//

	std::vector<float> e1Coefficients() const { return _e1Coef; }
	std::vector<float> e2Coefficients() const { return _e2Coef; }
	std::vector<float> e3Coefficients() const { return _e3Coef; }
	std::vector<float> e4Coefficients() const { return _e4Coef; }
		//
		//	PURPOSE:
		//		Return coefficient vectors driving the energy combination
		//		algorithm
		//


	~EmosEnergyCorrector() {}
		//
		//	PURPOSE:
		//		Destroy object.
		//

private:
	static std::vector<float>	_e1Coef, _e2Coef, _e3Coef, _e4Coef;
	static bool					_coefRead;
	float						_offsetGain[3][4][2];
	float						_fastG;
};


//
//	a version of the EnergyCorrector specialized for EPN
// 
class EpnEnergyCorrector : public virtual EnergyCorrector {
public:
	EpnEnergyCorrector(Ccd ccd,
					   const CcfConstituentTable &aduConv,
					   const Table *gainHigh,
					   const Table *reEmission,
					   const Ccf *ccf,
					   CalAtomCollection<EnergyCorrector> *parent = 0);
		//
		//	PURPOSE:
		//		Construct EpnEnergyCorrector object
		//
		//	ARGUMENTS:
		//		ccd		: CCD number this EnergyCorrector is applicable to
		//		aduConv	: CCF AduConv table from which to initialize object
		//		gainHigh: table with data driving high-accuracy gain algorithm
		//		reEmission: table with re-emission parameters
		//		ccf		: CCF in use
		//


	CalReal32Vector &gainCorrect(const CalReal64Vector &time,
								 const CalInt16Vector &rawX,
								 const CalInt16Vector &rawY,
								 const CalReal32Vector &eIn,
								 CalReal32Vector &eOut) const;
		//
		//	PURPOSE:
		//		Perform high-accuracy gain correction
		//
		//	ARGUMENTS:
		//		time	: vector with time tags
		//		rawX	: vector with rawX coordinates of events
		//		rawY	: vector with rawY coordinates of events
		//		eIn		: vector with uncorrected energies in ADU
		//		eOut	: vector with gain corrected energies in ADU
		//


	std::vector<double> additionalGainCtiCorrect(
											CalReal32Vector &ampl,
											const CalInt16Vector &rawY,
											double T,
											bool doTCorr = true,
											bool doOffsetCorr = true,
											bool doLongTermCTISoftCorr = true,
											bool doLongTermCTIyCorr = true,
											bool doLongTermCTICorr = true)
	const;
		//
		//	PURPOSE:
		//		Perform a sequence of additional CTI/gain related corrections
		//		that have to be applied after the principial correction for
		//		gain and CTI effects. The list of corrections is:
		//			o Orsay gain scaling [always]
		//			o long-term CTE decrease (modelled as gain) [optional]
		//			o CCD offsets [optional]
		//			o temperature dependent gain scaling [optional]
		//
		//	ARGUMENTS:
		//		ampl	: vector with input energy values; updated upon exit
		//		rawY	: vector with raw Y coordinates [1, 200] - needed for
		//				  long-term CTI correction
		//		T		: temperature [deg C] for gain correction
		//		doTCorr	:
		//		doOffsetCorr:
		//		doLongTermCTISoftCorr:
		//		doLongTermCTIyCorr:
		//		doLongTermCTICorr:
		//				  boolean switches to selectively enable (d/f) or
		//				  disable corrections provided by this method
		//
		//	RETURNS:
		//		parameter vector with data describing applied corrections -
		//		elements:
		//			0	: Orsay/in-orbit gain scaling factor
		//			1	: years elapsed since 1.1.2000
		//			2	: reference channel used for gain/CTI corrections
		//			3	: effect of long-term CTI increase
		//			4	: offset of individual CCD
		//			5	: temperature/gain-relation offset
		//			6	: temperature/gain-relation slope
		//			7	: temperature/gain-relation temperature
		//			8	: CTI increase energy function: term order 0
		//			9	: CTI increase energy function: term order 1
		//			10	: CTI increase energy function: term order 2
		//			11	: CTI increase reference Y position
		//


	void spatialCtiCorrect(CalReal32Vector &ampl,
                               const CalInt16Vector &rawX,
                               const CalInt16Vector &rawY) const;
		//
		//	PURPOSE:
		//		Perform an extra CTI correction based on
		//              fitting of a polynomial to each column of
		//		each CCD
		//
		//	ARGUMENTS:
		//		ampl	: vector with input energies in adu; updated upon exit
		//		rawX	: vector with column values
		//		rawY	: vector with row values
		//

	void spatialCtiCorrectSD20(CalReal32Vector &ampl,
                               const CalInt16Vector &rawX,
                               const CalInt16Vector &rawY) const;
		//
		//	PURPOSE:
		//		Perform an extra CTI correction based on
		//              spatial values calculated in Sanders et al. 2020
		//
		//	ARGUMENTS:
		//		ampl	: vector with input energies in adu; updated upon exit
		//		rawX	: vector with column values
		//		rawY	: vector with row values
	void combinedEventCorrect(CalReal32Vector &ampl,
                               const CalInt16Vector &pattId) const;
		//
		//	PURPOSE:
		//		Perform a CTI correction based on the
		//              pixel combination in the event
		//
		//	ARGUMENTS:
		//		ampl	: vector with input energies in adu; updated upon exit
		//		pattId  : vector with pattern values
		//


	void gainTimingCorrect(CalReal32Vector &ampl) const;
		//
		//	PURPOSE:
		//		Perform fine-adjustment of the gain in Timing mode; to be
		//		applied after all the principal gain corrections or
		//		immediatelty before the CTI corrections
		//
		//	ARGUMENTS:
		//		ampl	: vector with input energies in adu; updated upon
		//				  exit
		//


	void gainBurstCorrect(CalReal32Vector &ampl) const;
		//
		//	PURPOSE:
		//		Perform fine-adjustment of the gain in Burst mode; to be
		//		applied after all the principal gain corrections or
		//		immediatelty before the CTI corrections
		//
		//	ARGUMENTS:
		//		ampl	: vector with input energies in adu; updated upon
		//				  exit
		//


	void gainEffCorrect(CalReal32Vector &ampl) const;
		//
		//	PURPOSE:
		//		Perform fine-adjustment of the gain in EFF mode; to be
		//		applied after all the principal gain corrections or
		//		immediatelty before the CTI corrections
		//
		//	ARGUMENTS:
		//		ampl	: vector with input energies in adu; updated upon
		//				  exit
		//


	void gainFfCorrect(CalReal32Vector &ampl) const;
		//
		//	PURPOSE:
		//		Perform fine-adjustment of the gain in FF mode; to be
		//		applied after all the principal gain corrections or
		//		immediatelty before the CTI corrections
		//
		//	ARGUMENTS:
		//		ampl	: vector with input energies in adu; updated upon
		//				  exit
		//

	void bkgDepGainCorrect(CalReal32Vector &ampl,
			       int32 ndsclines, int16 mipsel,
			       int16 maxmip) const;
		//
		//	PURPOSE:
		//		Perform fine-adjustment of the gain based on the background
		//		rate, which is got from the number of discarded lines
		//
		//	ARGUMENTS:
		//		ampl	: vector with input energies in adu; updated upon
		//				  exit
	        //      ndsclines : the number of discarded lines
	        //      mipsel  : the number of lines either side of a MIP which are discarded
	        //      maxmip  : the maximum MIP value 
		//


	double reEmissionThreshold(double E, int16 rawY) const;
		//
		//	PURPOSE:
		//		Compute re-emission threshold value
		//
		//	ARGUMENTS:
		//		E	: energy [ADU]
		//		rawY: rawY coordinate of event
		//

	std::vector<float> getRateDepGainCoeffs() const;
		//
		//	PURPOSE:
		//		Returns the rate-dependent CTI coeffs.
		//

	std::vector<double> getRateDepPhaCoeffs(double Ne, bool withGain) const;
		//
		//	PURPOSE:
		//		Returns rate-dependent PHA coefficients dependent on input params
		//
		//	ARGUMENTS:
		//      Ne	 : The number of shifted electrons
		//	withGain : Use the with Gain coeffs or without coeffs  

        void readRDPHAcoeffs(Table *tab);
                //
                //      PURPOSE:
                //              Sets internal RDPHA vectors dependent on mode
                //                                 
	~EpnEnergyCorrector();
		//
		//	PURPOSE:
		//		Destroy object
		//

private:
	typedef std::vector<float *> EpnGainParameters;
	static const unsigned N_PARAMETERS = 10;
		//	this is a vector with the elements [0..9]:
		//		0    1    2    3    4     5     6   7     8   9
		//		pp0, pp1, pp2, pp3, dum1, dum2, aa, dum3, bb, dum

	static bool					_gainParamsSet;
	static EpnGainParameters	_gainParams;
	static std::vector<float>	_gainParamsSuppl;

	static std::vector<double>	_Acoef, _Bcoef;
	static double				_adu_eV, _rawy0;

	//	parameter for additional gain/CTI corrections
	std::vector<float>			//	Orsay-gain scaling; one value per CCD
	  _orsayGain,
	  
	  //	coefficients in modelling of long-term
	  //	CTE degradation:
	  //	time-dependence (polynomial)
	  _ccdOffsets,	//	one value per CCD [ADU]
	  _TgainOffset;	//	one value per CCD
	  /*It is better to use a map a ask for the value at a certain ccd */
	std::map<int,float>
	  _TgainSlope,	//	one value per CCD
	  _CRef;        //   correction factor to adjust peak position. One value per CCD and mode
	  /* This reference temperature no longer needed. Now, they want to use a constant (AI)*/
	  //_TNominal;	//	Nominal Quadrant Box Temperature per CCD (AI)


    std::vector<float>    _rdc_a;  // The rate-dependent CTI coeffs

   // Rate-dependent variables
    std::vector<std::vector<double> > _rdp_pha;  // The rate-dependent PHA constants

    std::vector<Slatec::LinearInterpolant<double> *> _rdp_corry_1, 
    			  _rdp_corry_2, _rdp_corry_3,
    			  _rdp_corrn_1,_rdp_corrn_2,_rdp_corrn_3,
    			  _rdp_ecorry_1,_rdp_ecorry_2,_rdp_ecorry_3,
    			  _rdp_ecorrn_1,_rdp_ecorrn_2,_rdp_ecorrn_3;
    bool _rdp_burst_support;

    std::vector<std::vector<float> >	_longTermCtiTcoef,
			//	energy-dependence
					_longTermCtiEcoef,
			//	energies 
					_longTermCtiEnergy,
			//	'shift factor'
					_longTermCtiShift;
    //	times
    std::vector<float> _longTermCtiTimes;

    double _longEBASE; // log of energy containing the base energy
			 // for non-interpolated long-term CTI correction

                                // Relevant CCD
    std::vector<short>			_longTermCtiCcd;

    int	_longAlgoid; // The algoid for the long-term CTI correction

    // Arrays for the discarded line correction
    std::vector<short>    _ndsclinCCD; //CCD number
    std::vector<float>	   _ndsclinOffset; // Offset
    std::vector<float>	   _ndsclinSlope; // Slope
    double _ndsclinMip2Conv;           // Conversion factor for MIPSEL=2 lines
    double _ndsclinMaxMip5Conv;        // Conversion factor for MAXMIP=5 lines


    // Arrays for the spatial correction
    std::vector<std::vector<float> >	_spatCtiPar,
                                        _spatCtiTmpl;
    std::vector<short>                  _spatCtiCcd;

    //std::vector<std::vector<float> <std::vector<float> >>
      //                                  _wdelta;
    // One element per CCD
    std::vector<SImage> 		_wdelta;
											 //  Spatial CTI block exists in CCF
    bool 				_spCtiRead;
											 //  EPN Spatial CTI CCF exists
    bool 				_spCtiSD20Read;

   // Energy interpolator for the combined event offsets
    std::vector<short>    _combEvCcd, // CCD number
                          _combEvPatt; // Pattern
    std::vector<float>	  _combEvEnergy, // Energy
                          _combEvShift; // Shift in ADU
    short                 _combEvMaxPatt; // Maximum pattern value

	float						_aduRef;		//	reference ADU number
	bool					 	_effGainRead;  	//  true if EFF gain corr.
												//  block exists in CCF
	bool					 	_ffGainRead;  	//  true if FF gain corr.
												//  block exists in CCF
	bool						_ndsclinGainRead; // true if NDSCLIN_GAIN
												// block exists in CCF
	bool						_combEvRead;    // true if COMB_EVT_OFFSET
											    // block exists in CCF

	struct {
		float	a1, a2,
				x0, dx;
	} _timingGain[4];	//	parameter for fine-gain adjustments in Timing (0)
						//	Burst (1), ExtendedFullFrame (2) and FF (3) mode
};


class RgsEnergyCorrector : public virtual EnergyCorrector {
public:
	RgsEnergyCorrector(Ccd ccd,
					   const CcfConstituentTable &aduConv,
					   const Table *darkFrameTable,
					   const Table *offsetOverrideTable,
					   CalAtomCollection<EnergyCorrector> *parent = 0);
		//
		//	PURPOSE:
		//		Construct RgsEnergyCorrector object
		//


	CalReal32Vector &offsetCorrect(const CalReal64Vector &time,
								   const CalInt16Vector &rawX,
								   const CalInt16Vector &rawY,
								   const CalInt8Vector &shape,
								   const CalInt16Vector &eIn,
								   CalReal32Vector &eOut) const;
	CalReal32Vector &offsetCorrect(const CalReal64Vector &time,
								   const CalInt16Vector &rawX,
								   const CalInt16Vector &rawY,
								   const CalInt8Vector &shape,
								   const CalInt8Vector &nodes,
								   const CalInt16Vector &eIn,
								   CalReal32Vector &eOut) const;
	CalReal32Vector &offsetCorrect(const CalReal64Vector &time,
								   const CalInt16Vector &rawX,
								   const CalInt8Vector &nodes,
								   const CalInt16Vector &eIn,
								   CalReal32Vector &eOut) const;
		//
		//	PURPOSE:
		//		Correct a vector of PHA channels for the effect of offset;
		//		third variant is for HTR mode where rawY/shape is meaningless
		//
		//	ARGUMENTS:
		//		time		: vector with time tags
		//		rawX/rawY	: pixel coordinates in PIXCOORD1
		//		shape		: number of pixels combined during on-board split
		//					  event reconstruction
		//		nodes		: node vector; for the first from, the node
		//					  is assumed to be fixed and taken from the
		//					  corresponding CAL state variable
		//		eIn			: to be corrected PHA values; conversion to
		//					  floating point type is done prior to perform
		//					  the correction; this optionally takes into
		//					  account randomization; see [2] for details
		//		eOut		: corrected values to be stored in this vector
		//


	float darkCurrent(int16 rawX, int16 rawY, unsigned node) const;
		//
		//	PURPOSE:
		//		Compute dark current at pixel position (rawX/rawY) for node
		//		<node>
		//

	static const unsigned NGRADES = 4;
	unsigned nGrades() const { return NGRADES; }
	std::vector<float> qSplitRatios() const;
	double rejectionThreshold(unsigned grade) const;
	double rejectionThreshold(unsigned grade, Node node) const;
	double acceptanceThreshold(unsigned grade) const;
	double acceptanceThreshold(unsigned grade, Node node) const;
		//
		//	PURPOSE:	
		//		Return rejection/acceptance threshold values in PI channels
		//		corresponding to a specified grade/node; in the second forms
		//		the node argument overrides the setting of the corresponding
		//		state variable
		//
		//	ARGUMENTS:
		//		grade:	number of pixels the event consists of
		//					1 <= grade <= 4
		//		node :	readout node identifier
		//					0 <= node <= 1


	~RgsEnergyCorrector();
		//
		//	PURPOSE:
		//		Destroy RGS energy corrector object
		//

protected:
	//	data members
	std::vector<double>	_rejectionThresholds,
					_acceptanceThresholds,
					_a0, _a1, _a2, _a3,
					_b0, _b1, _b2, _b3;			//	dark frame parameters
	double			_qSplitR[NGRADES];

	bool			_hasDiagnosticOffsetData;
	double			_eRandMin, _eRandMax;
	Slatec::CubicHermiteInterpolant<double>	*_offsetCurves[2];

private:
	void flatOffsetCurve(unsigned node);
	void flatOffsetCurves();
	void randomize(const CalInt16Vector &eIn, CalReal32Vector &eOut) const;
	unsigned short shapeToGrade(unsigned short shape) const;
};

#endif
