// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	MiscDataServer.h --- class to retrieve miscellaneous CCF data
//
//	DESCRIPTION:
//		A class MiscDataServer is declared instance of which are to be
//		used to retrive single-valued data from CCF::MiscData.
//		The query is keyword based.
//
//	CHANGE HISTORY:
//		UL, 1997-10-25: 1.0.0: - skeleton
//		UL, 1999-09-08: 2.0.0: - overhauled from CalAtomsXmm
//		UL, 2002-06-10: 2.1.0: - added has/getWithCcdNumber()
//
//	AUTHOR:
//		U. Lammers, Mon Oct  6 15:32:57 WET DST 1997
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /sasbuild/sas_git_repo/cvs_copy/sasdev/sas/packages/cal/src/MiscDataServer.h,v 1.11 2003/10/03 14:15:41 sasbuild Exp $
//

#ifndef MISC_DATA_SERVER_H
#define MISC_DATA_SERVER_H

#include <map>

#include "Ccf.h"
#include "CalChild.h"
#include "CalAtomsBase.h"

class MiscDataValue;	//	forward declaration
typedef std::map<std::string, const MiscDataValue *, std::less<std::string> >
	MiscDataMapType;

//
//	---( MiscDataServer )---------------------------------------------------
//
class MiscDataServer : public CalAtom,
				   public virtual CalChild<CalAtomCollection<MiscDataServer> >,
				   public MiscDataMapType {
public:
	MiscDataServer(const std::string &tableName,
				   CalAtomCollection<MiscDataServer> *parent = 0)
		: CalChild<CalAtomCollection<MiscDataServer> >(parent),
		  _tableName(tableName) {}
		//
		//	PURPOSE:
		//		Instantiate a MiscDataServer object
		//


	static unsigned id() { return Ccf::MiscData; }
		//
		//	PURPOSE:
		//		Return a uniue ID which for mapping Cal atoms onto
		//		CCF constituents
		//


	virtual void dump(std::ostream &os);
		//
		//	PURPOSE:
		//		Dump data members to an ostream
		//
		//	ARGUMENTS:
		//		In:
		//			os:	stream to print on
		//


	virtual bool has(const std::string &keyword) const;
	virtual bool hasWithCcdNumber(const std::string &keywordStem) const;
		//
		//	PURPOSE:
		//		Check whether a data item with the specified keyword
		//		exists
		//


	virtual double get(const std::string &keyword) const;
		//
		//	PURPOSE:
		//		Retrieve the value of the miscellaneous data item
		//		with the name <keyword>; Please note: attempting
		//		to retrieve a non-existing data item will result in
		//		an error condition.


	virtual double getWithCcdNumber(const std::string &keywordStem) const;
		//
		//	PURPOSE:
		//		Retrieve the value of the miscellaneous data item
		//		with the name <keyword>_ccd with <ccd> being the CCD
		//		number from the CAL state; Please note: attempting
		//		to retrieve a non-existing data item will result in
		//		an error condition.


	virtual ~MiscDataServer();
		//
		//	PURPOSE:
		//		Destrory MiscDataServer object constructed with above
		//		constructor(s).
		//

private:
	double getValueFromOtherConstituent(const std::string &) const;
	string stemCcd(const std::string &stem) const;

	std::string	_tableName;
};
	
#endif


