// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	OmFlatFieldServer.cc --- implementation of OmFlatFieldServer class
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 2000-01-26: 1.0.0: - initial version
//		UL, 2000-06-23: 1.1.0: - bug fix in OmFlatFiedServer constructor
//		UL, 2000-07-27: 1.2.0: - added retrieval of LED template maps
//		UL, 2001-05-04: 1.4.0: - access magnifier filter table as
//								 "FILTER-MAGNI" (see also SPR 2002)
//
//	AUTHOR:
//		U. Lammers (UL), Wed Jan 26 12:57:28 MET 2000
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Calibration Access and Data Handbook, XMM instrument team
//
//	$Header: /sasbuild/sas_git_repo/cvs_copy/sasdev/sas/packages/cal/src/OmFlatFieldServer.cc,v 1.17 2009/07/06 14:25:19 aibarra Exp $
//

#include "OmFlatFieldServer.h"
#include "CalOalUtility.h"

using namespace std;

//
//	---( CalAtomCollection<OmFlatFieldServer> ) ----------------------
//
/* AI gcc4 porting
 * too few template-parameter-lists => template<>
 */
template<>
CalAtomCollection<OmFlatFieldServer>::CalAtomCollection(
								const XmmInstrument *instrument,
								const Ccf *ccf,
								Xmm::Filter filter,
								AccuracyLevel accuracy)
{
	_filter		= filter;
	_accuracy	= accuracy;
	_nCCDs		= instrument->nCcds();
	_nNodes		= instrument->nNodes();
	_nEntries	= _nCCDs;
	OmFlatFieldServer *null = 0;
	_collection = new vector<OmFlatFieldServer *>(_nEntries, null);
	//
	//	read large-scale flatfield component from CCF
	//
	CifEntry *e = ccf->entry(instrument, Ccf::LargeScaleSens);
	DataSet *dataSet = e->open();
	string filterName = uppercase(Xmm::toString(filter));
	if (filter == Xmm::MAGNI)
		//
		//	OM Magnifier filter is referred to as "MAGNI" in [2] + [3]
		//	but we like to continue using the long name in the SAS;
		//	since [3] is frozen this is only possible with this ugly
		//	construct
		//
		filterName.erase(5); // "MAGNIFIER" -> "MAGNI"
	CcfConstituentArray ccfarr(dataSet, "FILTER-" +	filterName);
	const Array *array = ccfarr.block();
	//
	//	read attributes and coarse flatfield
	//
	double crvalx = array->attribute("CRVAL1")->asReal(),
		   crvaly = array->attribute("CRVAL2")->asReal(),
		   crpixx = array->attribute("CRPIX1")->asReal(),
		   crpixy = array->attribute("CRPIX2")->asReal(),
		   cdeltx = array->attribute("CDELT1")->asReal(),
		   cdelty = array->attribute("CDELT2")->asReal();
	vector<unsigned long> extents = array->size();
	ArrayData *ad = array->data();
	SImage lsmap(extents[0], extents[1], ad->real32Data());
	delete ad;

	ValueError lsaverage;
	lsaverage.value = array->attribute("AVERAGE")->asReal();
	lsaverage.error = array->attribute("DAVERAGE")->asReal();
	double lspixErr = array->attribute("PIXERR")->asReal();
	dataSetServer->close(dataSet);
	//
	//	read small-scale flatfield
	//
	e = ccf->entry(instrument, Ccf::PixToPixSens);
	dataSet = e->open();
	array = dataSet->array(e->queryBlockName("PIXTOPIXSENS"));
	extents = array->size();
	ad = array->data();
	SImage ssmap(extents[0], extents[1], ad->real32Data());
	delete ad;
	ValueError ssaverage;
	ssaverage.value = array->attribute("AVERAGE")->asReal();
	ssaverage.error = array->attribute("DAVERAGE")->asReal();
	double sspixErr = array->attribute("PIXERR")->asReal();

	//
	//	read LED template map
	//
	array = dataSet->array("LEDTEMPLATE");
	extents = array->size();
	ad = array->data();
	SImage ledmap(extents[0], extents[1], ad->real32Data());
	delete ad;

	ValueError ledaverage;
	ledaverage.value = array->attribute("AVERAGE")->asReal();
	ledaverage.error = array->attribute("DAVERAGE")->asReal();
	double ledpixErr = array->attribute("PIXERR")->asReal();

	dataSetServer->close(dataSet);
	//
	//	construct new OmFlatFieldServer
	//
	OmFlatFieldServer *&ff = (*_collection)[0];
	ff = new OmFlatFieldServer(crvalx, crpixx, cdeltx,
							   crvaly, crpixy, cdelty,
							   lsmap, lsaverage, lspixErr,
							   ssmap, ssaverage, sspixErr,
							   ledmap, ledaverage, ledpixErr,
							   this);
} // CalAtomCollection<OmFlatFieldServer>::CalAtomCollection()

/* AI gcc4 porting
 * too few template-parameter-lists => template<>
 */
template<>
OmFlatFieldServer *CalAtomCollection<OmFlatFieldServer>::select(
					Ccd,
					Node,
					unsigned) const
{
	return (*_collection)[0];
} // CalAtomCollection<OmFlatFieldServer>::select


//
//	---( OmFlatFieldServer> ) ---------------------------------------
//
OmFlatFieldServer::OmFlatFieldServer(
							   double crvalx, double crpixx, double cdeltx,
							   double crvaly, double crpixy, double cdelty,

							   const SImage &largeScaleFlatField,
							   ValueError largeScaleFlatFieldAverage,
							   double largeScalePixelError,

							   const SImage &smallScaleFlatField,
							   ValueError smallScaleFlatFieldAverage,
							   double smallScalePixelError,

							   const SImage &ledTemplateMap,
							   ValueError ledTemplateAverage,
							   double ledTemplatePixelError,
					CalAtomCollection<OmFlatFieldServer> *parent)
	: CalChild<CalAtomCollection<OmFlatFieldServer> >(parent),
		_largeScaleFlatFieldAverage(largeScaleFlatFieldAverage),
		_largeScalePixelError(largeScalePixelError),
		_smallScaleFlatFieldAverage(smallScaleFlatFieldAverage),
		_smallScalePixelError(smallScalePixelError),
		_ledTemplateAverage(ledTemplateAverage),
		_ledTemplatePixelError(ledTemplatePixelError)
{
	unsigned xExtent = largeScaleFlatField.xExtent(),
			 yExtent = largeScaleFlatField.yExtent();
	vector<float>	xd(xExtent),
					yd(yExtent);

	for (unsigned n=1; n<=xExtent; ++n)
		xd[n-1] = crvalx + (DBL(n)-crpixx) * cdeltx;
	for (unsigned n=1; n<=yExtent; ++n)
		yd[n-1] = crvaly + (DBL(n)-crpixy) * cdelty;

	vector<float> zd(largeScaleFlatField.data(),
					 largeScaleFlatField.data()+largeScaleFlatField.size());

	_largeScaleFlatField =
		new Slatec::Bivariate2DInterpolantRectangular<float>(xd, yd, zd);

	_smallScaleFlatField = new SImage(smallScaleFlatField);
	_ledTemplateMap = new SImage(ledTemplateMap);
} // OmFlatFieldServer::OmFlatFieldServer()


void OmFlatFieldServer::dump(ostream &os)
{
	os<<"OmFlatFieldServer{\n";
	os<<"}";
} // OmFlatFieldServer::dump()


SImage *OmFlatFieldServer::largeScaleFlatFieldMap(unsigned x0, unsigned y0,
												  unsigned dx, unsigned dy,
												  double)
const
{
	errstr<<"Warning: wavelength argument in large-scale flatfield "
			"retrieval currently ignored"<<Sas::warning("ArgumentIgnored");

	SImage *ffp = new SImage(dx, dy);
	SImage &ff = *ffp;

	for (unsigned nx=0; nx<dx; ++nx)
		for (unsigned ny=0; ny<dy; ++ny)
			ff[nx][ny] = (*_largeScaleFlatField)(FLT(x0+nx), FLT(y0+ny));

	return ffp;
} // OmFlatFieldServer::largeScaleFlatFieldMap()


ValueError OmFlatFieldServer::largeScaleFlatFieldAverage() const
{
	return _largeScaleFlatFieldAverage;
} // OmFlatFieldServer::largeScaleAverageFlatField()


double OmFlatFieldServer::largeScalePixelError() const
{
	return _largeScalePixelError;
} // OmFlatFieldServer::largeScalePixelError()


SImage *OmFlatFieldServer::smallScaleFlatFieldMap(unsigned x0, unsigned y0,
												  unsigned dx, unsigned dy,
												  double)
const
{
	errstr<<"Warning: countRateFrameRatio argument in small-scale flatfield "
			"retrieval currently ignored"<<Sas::warning("ArgumentIgnored");

	return _smallScaleFlatField->window(x0, y0, dx, dy);
} // OmFlatFieldServer::smallScaleFlatFieldMap()


ValueError OmFlatFieldServer::smallScaleFlatFieldAverage() const
{
	return _smallScaleFlatFieldAverage;
} // OmFlatFieldServer::smallScaleAverageFlatField()


double OmFlatFieldServer::smallScalePixelError() const
{
	return _smallScalePixelError;
} // OmFlatFieldServer::smallScalePixelError()


SImage *OmFlatFieldServer::ledTemplateMap(unsigned x0, unsigned y0,
										  unsigned dx, unsigned dy) const
{
	return _ledTemplateMap->window(x0, y0, dx, dy);
} // OmFlatFieldServer::ledTemplateMap()


ValueError OmFlatFieldServer::ledTemplateAverage() const
{
	return _ledTemplateAverage;
} // OmFlatFieldServer::ledTemplateAverage()


double OmFlatFieldServer::ledTemplatePixelError() const
{
	return _ledTemplatePixelError;
} // OmFlatFieldServer::ledTemplatePixelError()


OmFlatFieldServer::~OmFlatFieldServer()
{
	delete _largeScaleFlatField; _largeScaleFlatField = 0;
	delete _smallScaleFlatField; _smallScaleFlatField = 0;
	delete _ledTemplateMap; _ledTemplateMap = 0;
} // OmFlatFieldServer::~OmFlatFieldServer()
