// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	RgsSolarAspectAngServer --- implementation of atom to retreive 
//				    solar aspect angle coefficients
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		RDS, 2013-02-26, 1.0.0: 
//
//	AUTHOR:
//		R. Saxton
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//
//	$Header: /sasbuild/sas_git_repo/cvs_copy/sasdev/sas/packages/cal/src/RgsSolarAspectAngServer.cc,v 1.2 2013/02/27 16:24:49 rsaxton Exp $
//

#include <functional>	//	binders
#include <algorithm>
#include <cmath>

#include "CalOalUtility.h"
#include "CalServer.h"
#include "RgsSolarAspectAngServer.h"

using namespace CalOalUtility;
using namespace std;

//
//	------------------------( RgsSolarAspectAngServer )-----------------------------
//

RgsSolarAspectAngServer::RgsSolarAspectAngServer(vector<double> saacoeff_order1,
                                                 vector<double> saacoeff_order2,
                                                 CalAtomCollection<RgsSolarAspectAngServer> *parent)
	: CalChild<CalAtomCollection<RgsSolarAspectAngServer> >(parent),
	  _saacoeff_order1(saacoeff_order1), _saacoeff_order2(saacoeff_order2)
{
} // RgsSolarAspectAngServer::RgsSolarAspectAngServer()


vector<double> RgsSolarAspectAngServer::coefficients(unsigned order) const
{
        if (order != 1 && order != 2)
           errstr<<"Order "<<order<<" not recognised"
                 <<Sas::error("UnknownOrder");

        if (order == 1)
           return _saacoeff_order1;
        else // order=2
           return _saacoeff_order2;

} // RgsSolarAspectAngServer::coefficients()


RgsSolarAspectAngServer::~RgsSolarAspectAngServer()
{
} // RgsSolarAspectAngServer::~RgsSolarAspectAngServer()

/* AI gcc4 porting
 * too few template-parameter-lists
 * => template<>
 */
template<>
CalAtomCollection<RgsSolarAspectAngServer>::CalAtomCollection(
								const XmmInstrument *instrument,
								const Ccf *ccf,
								Xmm::Filter filter,
								AccuracyLevel accuracy)
{
	_filter		= filter;
	_accuracy	= accuracy;
        _nEntries       = 1; 
	RgsSolarAspectAngServer	*null = 0;
	_collection	= new vector<RgsSolarAspectAngServer *>(_nEntries, 
                                                                      null);

       // Read the Solar aspect angle coefficients
	CifEntry *e = ccf->entry(instrument, Ccf::SAACorr);
	DataSet *dataSet = e->open();
	CcfConstituentTable ccftab(dataSet, e->queryBlockName(
                                    Ccf::toString(Ccf::SAACorr)));
	const Table *table = ccftab.block();

	ColumnData *cdorder = table->column("DISPORDER")->data(),
			   *cdpara = table->column("PARA")->data(),
			   *cdparb = table->column("PARB")->data();
        int8 *order = cdorder->int8Data();
        real32 *para = cdpara->real32Data(),
               *parb = cdparb->real32Data();

       // Write data into the Solar aspect angle server 
        vector<double> saacoeff_order1, saacoeff_order2;

	for (unsigned n=0; n<table->rows(); ++n) {

           // Use discrete vector for each order as only 2
            if (order[n] == 1) {
               saacoeff_order1.push_back(para[n]);
               saacoeff_order1.push_back(parb[n]);
            } else { // Order 2
               saacoeff_order2.push_back(para[n]);
               saacoeff_order2.push_back(parb[n]);
            }
        }

	(*_collection)[0] = new RgsSolarAspectAngServer(saacoeff_order1,
                                                        saacoeff_order2);

	//	tidy up
	delete cdorder;
	delete cdpara;
	delete cdparb;

	dataSetServer->close(dataSet);
} // CalAtomCollection<RgsSolarAspectAngServer>::CalAtomCollection()

/* AI gcc4 porting
 * too few template-parameter-lists
 * => template<>
 */
template<>
RgsSolarAspectAngServer *CalAtomCollection<RgsSolarAspectAngServer>::select(
					Ccd,
					Node,
					unsigned) const
{
	return (*_collection)[0];
} // CalAtomCollection<RgsSolarAspectAngServer>::select


